@@ -1,9 +1,9 @@
 ;;; python-pytest.el --- helpers to run pytest -*- lexical-binding: t; -*-
 
 ;; Author: wouter bolsterlee <wouter@bolsterl.ee>
-;; Package-Version: 3.4.0
-;; Package-Revision: 3.4.0-0-g46fd00646225
-;; Package-Requires: ((emacs "24.4") (dash "2.18.0") (transient "0.3.7") (projectile "0.14.0") (s "1.12.0"))
+;; Package-Version: 20250726.1726
+;; Package-Revision: ed2ecee09d1c
+;; Package-Requires: ((emacs "24.4") (dash "2.18.0") (transient "0.3.7") (s "1.12.0"))
 ;; Keywords: pytest, test, python, languages, processes, tools
 ;; URL: https://github.com/wbolster/emacs-python-pytest
 ;;
@@ -26,9 +26,12 @@
 
 (require 'dash)
 (require 'transient)
-(require 'projectile)
 (require 's)
 
+(require 'projectile nil t)
+(require 'project nil t)
+(require 'treesit nil t)
+
 (defgroup python-pytest nil
   "pytest integration"
   :group 'python
@@ -50,6 +53,15 @@ When t, this toggles the behaviour of the prefix argument."
   :group 'python-pytest
   :type 'string)
 
+(defcustom python-pytest-shell-executable
+  (if (eq system-type 'windows-nt)
+      (or (executable-find "powershell")
+          (executable-find "cmdproxy"))
+    "sh")
+  "The name of the pytest shell executable."
+  :group 'python-pytest
+  :type 'string)
+
 (defcustom python-pytest-setup-hook nil
   "Hooks to run before a pytest process starts."
   :group 'python-pytest
@@ -107,6 +119,54 @@ When non-nil only ‘test_foo()’ will match, and nothing else."
                  (const :tag "Save current buffer" save-current)
                  (const :tag "Ignore" nil)))
 
+(defcustom python-pytest-preferred-project-manager 'auto
+  "Override `projectile' or `project' auto-discovery to set preference if using both."
+  :group 'python-pytest
+  :type '(choice (const :tag "Projectile" projectile)
+                 (const :tag "Project" project)
+                 (const :tag "Automatically selected" auto))
+  :set (lambda (symbol value)
+         (cond
+          ((and (eq value 'projectile)
+                (not (featurep 'projectile)))
+           (user-error "Projectile preferred for python-pytest.el, but not available."))
+          ((and (eq value 'project)
+                (not (fboundp 'project-root)))
+           (user-error (concat "Project.el preferred for python-pytest.el, "
+                               "but need a newer version of Project (28.1+) to use.")))
+          (t
+           (set-default symbol value)
+           value))))
+
+(defcustom python-pytest-project-root-override nil
+  "Override directory where pytest runs.
+It is intended to be set in monorepo which has multiple roots to run pytest,
+and to be set via .dir-locals.el."
+  :group 'python-pytest
+  :type '(choice directory
+                 (const :tag "No override" nil)))
+
+(defcustom python-pytest-project-name-override nil
+  "Override project name.
+It should be set in monorepo which has multiple roots to run pytest.
+Each root must have setting for `python-pytest-project-root-override'
+in .dir-locals.el.
+
+It is also intended to be set via .dir-locals.el."
+  :group 'python-pytest
+  :type '(choice directory
+                 (const :tag "No override" nil)))
+
+(defcustom python-pytest-use-treesit (featurep 'treesit)
+  "Whether to use treesit for getting the node ids of things at point.
+
+Users that are running a version of Emacs that supports treesit
+and have the Python language grammar for treesit should set this
+variable to t. Users that are running a version of Emacs that
+don't support treesit should set this variable to nil."
+  :group 'python-pytest
+  :type 'boolean)
+
 (defvar python-pytest--history nil
   "History for pytest invocations.")
 
@@ -126,7 +186,8 @@ When non-nil only ‘test_foo()’ will match, and nothing else."
    [("-c" "color" "--color")
     ("-q" "quiet" "--quiet")
     ("-s" "no output capture" "--capture=no")
-    (python-pytest:-v)]]
+    (python-pytest:-v)
+    (python-pytest:--l)]]
   ["Selection, filtering, ordering"
    [(python-pytest:-k)
     (python-pytest:-m)
@@ -157,8 +218,10 @@ When non-nil only ‘test_foo()’ will match, and nothing else."
     ("F" "file (this)" python-pytest-file)]
    [("m" "files" python-pytest-files)
     ("M" "directories" python-pytest-directories)]
-   [("d" "def/class (dwim)" python-pytest-function-dwim)
-    ("D" "def/class (this)" python-pytest-function)]])
+   [("d" "def at point (dwim)" python-pytest-run-def-or-class-at-point-dwim :if-not python-pytest--use-treesit-p)
+    ("D" "def at point" python-pytest-run-def-or-class-at-point :if-not python-pytest--use-treesit-p)
+    ("d" "def at point" python-pytest-run-def-at-point-treesit :if python-pytest--use-treesit-p)
+    ("c" "class at point" python-pytest-run-class-at-point-treesit :if python-pytest--use-treesit-p)]])
 
 (define-obsolete-function-alias 'python-pytest-popup 'python-pytest-dispatch "2.0.0")
 
@@ -237,7 +300,27 @@ With a prefix argument, allow editing."
    :edit current-prefix-arg))
 
 ;;;###autoload
-(defun python-pytest-function (file func args)
+(defun python-pytest-run-def-at-point-treesit ()
+  "Run def at point."
+  (interactive)
+  (python-pytest--run
+   :args (transient-args 'python-pytest-dispatch)
+   :file (buffer-file-name)
+   :node-id (python-pytest--node-id-def-at-point-treesit)
+   :edit current-prefix-arg))
+
+;;;###autoload
+(defun python-pytest-run-class-at-point-treesit ()
+  "Run class at point."
+  (interactive)
+  (python-pytest--run
+   :args (transient-args 'python-pytest-dispatch)
+   :file (buffer-file-name)
+   :node-id (python-pytest--node-id-class-at-point-treesit)
+   :edit current-prefix-arg))
+
+;;;###autoload
+(defun python-pytest-run-def-or-class-at-point (file func args)
   "Run pytest on FILE with FUNC (or class).
 
 Additional ARGS are passed along to pytest.
@@ -245,27 +328,32 @@ With a prefix argument, allow editing."
   (interactive
    (list
     (buffer-file-name)
-    (python-pytest--current-defun)
+    (python-pytest--node-id-def-or-class-at-point)
     (transient-args 'python-pytest-dispatch)))
   (python-pytest--run
    :args args
    :file file
-   :func func
+   :node-id func
    :edit current-prefix-arg))
 
 ;;;###autoload
-(defun python-pytest-function-dwim (file func args)
-  "Run pytest on FILE with FUNC (or class).
+(defun python-pytest-run-def-or-class-at-point-dwim (file func args)
+  "Run pytest on FILE using FUNC at point as the node-id.
 
-When run interactively, this tries to work sensibly using
-the current file and function around point.
+If `python-pytest--test-file-p' returns t for FILE (i.e. the file
+is a test file), then this function results in the same behavior
+as calling `python-pytest-run-def-at-point'. If
+`python-pytest--test-file-p' returns nil for FILE (i.e. the
+current file is not a test file), then this function will try to
+find related test files and test defs (i.e. sensible match) for
+the current file and the def at point.
 
 Additional ARGS are passed along to pytest.
 With a prefix argument, allow editing."
   (interactive
    (list
     (buffer-file-name)
-    (python-pytest--current-defun)
+    (python-pytest--node-id-def-or-class-at-point)
     (transient-args 'python-pytest-dispatch)))
   (unless (python-pytest--test-file-p file)
     (setq
@@ -292,7 +380,7 @@ With a prefix argument, allow editing."
   (python-pytest--run
    :args args
    :file file
-   :func func
+   :node-id func
    :edit current-prefix-arg))
 
 ;;;###autoload
@@ -339,16 +427,22 @@ With a prefix ARG, allow editing."
     map)
   "Keymap for `python-pytest-mode' major mode.")
 
-(cl-defun python-pytest--run (&key args file func edit)
-  "Run pytest for the given arguments."
+(cl-defun python-pytest--run (&key args file node-id edit)
+  "Run pytest for the given arguments.
+
+NODE-ID should be the node id of the test to run. pytest uses
+double colon \"::\" for separating components in node ids. For
+example, the node-id for a function outside a class is the
+function name, the node-id for a function inside a class is
+TestClass::test_my_function, the node-id for a function inside a
+class that is inside another class is
+TestClassParent::TestClassChild::test_my_function."
   (setq args (python-pytest--transform-arguments args))
   (when (and file (file-name-absolute-p file))
     (setq file (python-pytest--relative-file-name file)))
-  (when func
-    (setq func (s-replace "." "::" func)))
   (let ((command)
         (thing (cond
-                ((and file func) (format "%s::%s" file func))
+                ((and file node-id) (format "%s::%s" file node-id))
                 (file file))))
     (when thing
       (setq args (-snoc args (python-pytest--shell-quote thing))))
@@ -402,12 +496,23 @@ With a prefix ARG, allow editing."
       (run-hooks 'python-pytest-setup-hook)
       (make-comint-in-buffer
        "pytest" buffer
-       (if (eq system-type 'windows-nt) "cmdproxy" "sh")
+       python-pytest-shell-executable
        nil "-c" command)
       (run-hooks 'python-pytest-started-hook)
       (setq process (get-buffer-process buffer))
       (set-process-sentinel process #'python-pytest--process-sentinel))))
 
+(defun python-pytest--use-treesit-p ()
+  "Return t if python-pytest-use-treesit is t. Otherwise, return nil.
+
+This function is passed to the parameter :if in
+`python-pytest-dispatch'.
+
+Although this function might look useless, the main reason why it
+was defined was that the parameter that is provided to the
+transient keyword :if must be a function."
+  python-pytest-use-treesit)
+
 (defun python-pytest--shell-quote (s)
   "Quote S for use in a shell command. Like `shell-quote-argument', but prettier."
   (if (s-equals-p s (shell-quote-argument s))
@@ -464,6 +569,13 @@ When present ON-REPLACEMENT is substituted, else OFF-REPLACEMENT is appended."
          (formatted-input (format " %s" quoted-input)))
     formatted-input))
 
+(transient-define-argument python-pytest:--l ()
+  :description "set log cli level"
+  :class 'transient-option
+  :key "--l"
+  :argument "--log-cli-level="
+  :choices '("debug" "info" "warning" "error" "critical"))
+
 (transient-define-argument python-pytest:-k ()
   :description "only names matching expression"
   :class 'transient-option
@@ -509,10 +621,148 @@ When present ON-REPLACEMENT is substituted, else OFF-REPLACEMENT is appended."
   :argument "--numprocesses="
   :choices '("auto" "0" "1" "2" "4" "8" "16"))
 
+(defun python-pytest--using-projectile ()
+  "Returns t if projectile being used for project management."
+  (or (eq python-pytest-preferred-project-manager 'projectile)
+      (and (eq python-pytest-preferred-project-manager 'auto)
+           (bound-and-true-p projectile-mode))))
 
 ;; python helpers
 
-(defun python-pytest--current-defun ()
+(defun python-pytest--point-is-inside-def-treesit ()
+  (unless (treesit-language-available-p 'python)
+    (error "This function requires tree-sitter support for python, but it is not available."))
+  (save-restriction
+    (widen)
+    (catch 'return
+      (let ((current-node (treesit-node-at (point) 'python)))
+        (while (setq current-node (treesit-node-parent current-node))
+          (when (equal (treesit-node-type current-node) "function_definition")
+            (throw 'return t)))))))
+
+(defun python-pytest--point-is-inside-class-treesit ()
+  (unless (treesit-language-available-p 'python)
+    (error "This function requires tree-sitter support for python, but it is not available."))
+  (save-restriction
+    (widen)
+    (catch 'return
+      (let ((current-node (treesit-node-at (point) 'python)))
+        (while (setq current-node (treesit-node-parent current-node))
+          (when (equal (treesit-node-type current-node) "class_definition")
+            (throw 'return t)))))))
+
+(defun python-pytest--node-id-def-at-point-treesit ()
+  "Return the node id of the def at point.
+
++ If the test function is not inside a class, its node id is the name
+  of the function.
++ If the test function is defined inside a class, its node id would
+  look like: TestGroup::test_my_function.
++ If the test function is defined inside a class that is defined
+  inside another class, its node id would look like:
+  TestGroupParent::TestGroupChild::test_my_function."
+  (unless (python-pytest--point-is-inside-def-treesit)
+    (error "The point is not inside a def."))
+  (save-restriction
+    (widen)
+    (let ((function
+           ;; Move up to the outermost function
+           (catch 'return
+             (let ((current-node (treesit-node-at (point) 'python))
+                   function-node)
+               (catch 'break
+                 (while (setq current-node (treesit-node-parent current-node))
+                   (when (equal (treesit-node-type current-node) "function_definition")
+                     (setq function-node current-node)
+                     ;; At this point, we know that we are on a
+                     ;; function. We need to move up to see if the
+                     ;; function is inside a function. If that's the
+                     ;; case, we move up. This way, we find the
+                     ;; outermost function. We need to do this because
+                     ;; pytest can't execute functions inside functions,
+                     ;; so we must get the function that is not inside
+                     ;; other function.
+                     (while (setq current-node (treesit-node-parent current-node))
+                       (when (equal (treesit-node-type current-node) "function_definition")
+                         (setq function-node current-node)))
+                     (throw 'break nil))))
+               (dolist (child (treesit-node-children function-node))
+                 (when (equal (treesit-node-type child) "identifier")
+                   (throw 'return
+                          (cons
+                           ;; Keep a reference to the node that is a
+                           ;; function_definition. We need this
+                           ;; reference because afterwards we need to
+                           ;; move up starting at the current node to
+                           ;; find the node id of the class (if there's
+                           ;; any) in which the function is defined.
+                           function-node
+                           (buffer-substring-no-properties
+                            (treesit-node-start child)
+                            (treesit-node-end child)))))))))
+          parents)
+      ;; Move up through the parent nodes to see if the function is
+      ;; defined inside a class and collect the classes to finally build
+      ;; the node id of the current function. Remember that the node id
+      ;; of a function that is defined within nested classes must have
+      ;; the name of the nested classes.
+      (let ((current-node (car function)))
+        (while (setq current-node (treesit-node-parent current-node))
+          (when (equal (treesit-node-type current-node) "class_definition")
+            (dolist (child (treesit-node-children current-node))
+              (when (equal (treesit-node-type child) "identifier")
+                (push (buffer-substring-no-properties
+                       (treesit-node-start child)
+                       (treesit-node-end child))
+                      parents))))))
+      (string-join `(,@parents ,(cdr function)) "::"))))
+
+(defun python-pytest--node-id-class-at-point-treesit ()
+  "Return the node id of the class at point.
+
++ If the class is not inside another class, its node id is the name
+  of the class.
++ If the class is defined inside another class, the node id of the
+  class which is contained would be: TestGroupParent::TestGroupChild,
+  while the node id of the class which contains the other class would
+  be TestGroupParent."
+  (unless (python-pytest--point-is-inside-class-treesit)
+    (error "The point is not inside a class."))
+  (save-restriction
+    (widen)
+    (let ((class
+           ;; Move up to the outermost function
+           (catch 'return
+             (let ((current-node (treesit-node-at (point) 'python)))
+               (catch 'break
+                 (while (setq current-node (treesit-node-parent current-node))
+                   (when (equal (treesit-node-type current-node) "class_definition")
+                     (throw 'break nil))))
+               (dolist (child (treesit-node-children current-node))
+                 (when (equal (treesit-node-type child) "identifier")
+                   (throw 'return
+                          (cons
+                           ;; Keep a reference to the node that is a
+                           ;; function_definition
+                           current-node
+                           (buffer-substring-no-properties
+                            (treesit-node-start child)
+                            (treesit-node-end child)))))))))
+          parents)
+      ;; Move up through the parents to collect the list of classes in
+      ;; which the class is contained. pytest supports running nested
+      ;; classes, but it doesn't support runing nested functions.
+      (let ((current-node (car class)))
+        (while (setq current-node (treesit-node-parent current-node))
+          (when (equal (treesit-node-type current-node) "class_definition")
+            (dolist (child (treesit-node-children current-node))
+              (when (equal (treesit-node-type child) "identifier")
+                (push (buffer-substring-no-properties
+                       (treesit-node-start child)
+                       (treesit-node-end child))
+                      parents))))))
+      (string-join `(,@parents ,(cdr class)) "::"))))
+(defun python-pytest--node-id-def-or-class-at-point ()
   "Detect the current function/class (if any)."
   (let* ((name
           (or (python-info-current-defun)
@@ -532,7 +782,7 @@ When present ON-REPLACEMENT is substituted, else OFF-REPLACEMENT is appended."
           (if (s-lowercase? (substring name 0 1))
               (car (s-split-up-to "\\." name 1))
             name)))
-    name))
+    (s-replace "." "::" name)))
 
 (defun python-pytest--make-test-name (func)
   "Turn function name FUNC into a name (hopefully) matching its test name.
@@ -552,12 +802,29 @@ Example: ‘MyABCThingy.__repr__’ becomes ‘test_my_abc_thingy_repr’."
 
 (defun python-pytest--project-name ()
   "Find the project name."
-  (projectile-project-name))
+  (cond
+   (python-pytest-project-name-override)
+   ((python-pytest--using-projectile)
+    (projectile-project-name))
+   ((fboundp 'project-name)
+    (project-name (project-current)))
+   ;; older emacs...
+   (t (file-name-nondirectory
+       (directory-file-name (car (project-roots (project-current))))))))
 
 (defun python-pytest--project-root ()
-  "Find the project root directory."
-  (let ((projectile-require-project-root nil))
-    (projectile-compilation-dir)))
+  "Find the project root directory.
+The return value can manually set your own `python-pytest-project-root-override'
+variable."
+  (cond
+   (python-pytest-project-root-override)
+   ((python-pytest--using-projectile)
+    (let ((projectile-require-project-root nil))
+      (projectile-compilation-dir)))
+   ((fboundp 'project-root)
+    (project-root (project-current)))
+   ;; pre-emacs "28.1"
+   (t (car (project-roots (project-current))))))
 
 (defun python-pytest--relative-file-name (file)
   "Make FILE relative to the project root."
@@ -568,11 +835,29 @@ Example: ‘MyABCThingy.__repr__’ becomes ‘test_my_abc_thingy_repr’."
 
 (defun python-pytest--test-file-p (file)
   "Tell whether FILE is a test file."
-  (projectile-test-file-p file))
+  (if (python-pytest--using-projectile)
+      (projectile-test-file-p file)
+    (let ((base-name (file-name-nondirectory file)))
+      (or (string-prefix-p "test_" base-name)
+          (string-suffix-p "_test.py" base-name)))))
 
 (defun python-pytest--find-test-file (file)
   "Find a test file associated to FILE, if any."
-  (let ((test-file (projectile-find-matching-test file)))
+  (let ((test-file))
+    (if (python-pytest--using-projectile)
+        (setq test-file (projectile-find-matching-test file))
+      (let* ((base-name (file-name-sans-extension (file-name-nondirectory file)))
+             (test-file-regex (concat "\\`test_"
+                                      base-name "\\.py\\'\\|\\`"
+                                      base-name "_test\\.py\\'")))
+        (setq test-file
+              (car (cl-delete-if
+                    (lambda (full-file)
+                      (let ((file (file-name-nondirectory full-file)))
+                        (not (string-match-p
+                              test-file-regex
+                              file))))
+                    (project-files (project-current t)))))))
     (unless test-file
       (user-error "No test file found"))
     test-file))
@@ -585,36 +870,45 @@ Example: ‘MyABCThingy.__repr__’ becomes ‘test_my_abc_thingy_repr’."
 
 (cl-defun python-pytest--select-test-files (&key type)
   "Interactively choose test files."
-  (cl-block nil
-    (let* ((test-files
-            (->> (projectile-project-files (python-pytest--project-root))
-                 (-sort 'string<)
-                 (projectile-sort-by-recentf-first)
-                 ;; show test files if any found, otherwise show everything
-                 (funcall (-orfn #'projectile-test-files #'identity))))
-           (test-directories
-            (->> test-files
-                 (-map 'file-name-directory)
-                 (-uniq)
-                 (-sort 'string<)))
-           (candidates (if (eq type 'file) test-files test-directories))
-           (done-message (propertize "[finish test file selection]" 'face 'success))
-           (choices)
-           (choice)
-           (selection-active t))
-      (unless candidates
-        (user-error "No test files found"))
-      (while (and selection-active candidates)
-        (setq choice (completing-read
-                      "Choose test files: "
-                      (if choices (cons done-message candidates) candidates)
-                      nil t))
-        (if (s-equals-p choice done-message)
-            (setq selection-active nil)
-          (setq
-           choices (cons choice choices)
-           candidates (-remove-item choice candidates))))
-      (cl-return (reverse choices)))))
+  (let* ((test-files
+          (if (python-pytest--using-projectile)
+              (->> (projectile-project-files (python-pytest--project-root))
+                   (-sort 'string<)
+                   (projectile-sort-by-recentf-first)
+                   ;; show test files if any found, otherwise show everything
+                   (funcall (-orfn #'projectile-test-files #'identity)))
+            (let* ((vc-directory-exclusion-list
+                    (append vc-directory-exclusion-list '("venv" ".venv")))
+                   (sorted-test-files
+                    (sort (cl-delete-if
+                           (lambda (file)
+                             (not (python-pytest--test-file-p file)))
+                           (project-files (project-current t)))
+                          #'string<))
+                   (recentf-test-files '())
+                   (test-files-prj
+                    (when (fboundp 'recentf)
+                      (dolist (file recentf-list
+                                    (progn
+                                      (setq sorted-test-files
+                                            (append (nreverse recentf-test-files)
+                                                    sorted-test-files))
+                                      (cl-delete-duplicates sorted-test-files
+                                                            :test 'equal )))
+                        (when (and (file-exists-p file)
+                                   (python-pytest--test-file-p file))
+                          (push (expand-file-name file) recentf-test-files))))))
+              test-files-prj)))
+         (test-directories
+          (->> test-files
+               (-map 'file-name-directory)
+               (-uniq)
+               (-sort 'string<)))
+         (candidates (if (eq type 'file) test-files test-directories))
+         (prompt (if (eq type 'file) "Choose test files: " "Choose test directories: ")))
+    (unless candidates
+      (user-error "No test files or directories found"))
+    (completing-read-multiple prompt candidates nil t)))
 
 (defun python-pytest--maybe-save-buffers ()
   "Maybe save modified buffers."
@@ -630,7 +924,9 @@ Example: ‘MyABCThingy.__repr__’ becomes ‘test_my_abc_thingy_repr’."
     ;; check all project buffers
     (-when-let*
         ((buffers
-          (projectile-buffers-with-file (projectile-project-buffers)))
+          (if (python-pytest--using-projectile)
+              (projectile-buffers-with-file (projectile-project-buffers))
+            (-filter 'buffer-file-name (project-buffers (project-current t)))))
          (modified-buffers
           (-filter 'buffer-modified-p buffers))
          (confirmed
